from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    ModelTemplate,
    ModelTemplateWrite,
    ModelTemplateFields,
    ModelTemplateList,
    ModelTemplateWriteList,
    ModelTemplateTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._model_template import (
    _MODELTEMPLATE_PROPERTIES_BY_FIELD,
    _create_model_template_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .model_template_base_mappings import ModelTemplateBaseMappingsAPI
from .model_template_query import ModelTemplateQueryAPI


class ModelTemplateAPI(NodeAPI[ModelTemplate, ModelTemplateWrite, ModelTemplateList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[ModelTemplate]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ModelTemplate,
            class_list=ModelTemplateList,
            class_write_list=ModelTemplateWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.base_mappings_edge = ModelTemplateBaseMappingsAPI(client)

    def __call__(
        self,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> ModelTemplateQueryAPI[ModelTemplateList]:
        """Query starting at model templates.

        Args:
            version_: The version to filter on.
            version_prefix: The prefix of the version to filter on.
            shop_version: The shop version to filter on.
            shop_version_prefix: The prefix of the shop version to filter on.
            watercourse: The watercourse to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of model templates to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for model templates.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_model_template_filter(
            self._view_id,
            version_,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(ModelTemplateList)
        return ModelTemplateQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        model_template: ModelTemplateWrite | Sequence[ModelTemplateWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) model templates.

        Note: This method iterates through all nodes and timeseries linked to model_template and creates them including the edges
        between the nodes. For example, if any of `base_mappings` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            model_template: Model template or sequence of model templates to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new model_template:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import ModelTemplateWrite
                >>> client = PowerOpsModelsV1Client()
                >>> model_template = ModelTemplateWrite(external_id="my_model_template", ...)
                >>> result = client.model_template.apply(model_template)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.model_template.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(model_template, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more model template.

        Args:
            external_id: External id of the model template to delete.
            space: The space where all the model template are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete model_template by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.model_template.delete("my_model_template")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.model_template.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> ModelTemplate | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> ModelTemplateList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> ModelTemplate | ModelTemplateList | None:
        """Retrieve one or more model templates by id(s).

        Args:
            external_id: External id or list of external ids of the model templates.
            space: The space where all the model templates are located.

        Returns:
            The requested model templates.

        Examples:

            Retrieve model_template by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> model_template = client.model_template.retrieve("my_model_template")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.base_mappings_edge,
                    "base_mappings",
                    dm.DirectRelationReference("sp_powerops_types", "ModelTemplate.baseMappings"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Mapping", "1"),
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ModelTemplateList:
        """Search model templates

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            version_: The version to filter on.
            version_prefix: The prefix of the version to filter on.
            shop_version: The shop version to filter on.
            shop_version_prefix: The prefix of the shop version to filter on.
            watercourse: The watercourse to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of model templates to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results model templates matching the query.

        Examples:

           Search for 'my_model_template' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> model_templates = client.model_template.search('my_model_template')

        """
        filter_ = _create_model_template_filter(
            self._view_id,
            version_,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _MODELTEMPLATE_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: ModelTemplateFields | Sequence[ModelTemplateFields] = None,
        query: str | None = None,
        search_properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        query: str | None = None,
        search_property: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across model templates

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            version_: The version to filter on.
            version_prefix: The prefix of the version to filter on.
            shop_version: The shop version to filter on.
            shop_version_prefix: The prefix of the shop version to filter on.
            watercourse: The watercourse to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of model templates to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count model templates in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.model_template.aggregate("count", space="my_space")

        """

        filter_ = _create_model_template_filter(
            self._view_id,
            version_,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _MODELTEMPLATE_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ModelTemplateFields,
        interval: float,
        query: str | None = None,
        search_property: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for model templates

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            version_: The version to filter on.
            version_prefix: The prefix of the version to filter on.
            shop_version: The shop version to filter on.
            shop_version_prefix: The prefix of the shop version to filter on.
            watercourse: The watercourse to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of model templates to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_model_template_filter(
            self._view_id,
            version_,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _MODELTEMPLATE_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        version_: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ModelTemplateList:
        """List/filter model templates

        Args:
            version_: The version to filter on.
            version_prefix: The prefix of the version to filter on.
            shop_version: The shop version to filter on.
            shop_version_prefix: The prefix of the shop version to filter on.
            watercourse: The watercourse to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of model templates to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `base_mappings` external ids for the model templates. Defaults to True.

        Returns:
            List of requested model templates

        Examples:

            List model templates and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> model_templates = client.model_template.list(limit=5)

        """
        filter_ = _create_model_template_filter(
            self._view_id,
            version_,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.base_mappings_edge,
                    "base_mappings",
                    dm.DirectRelationReference("sp_powerops_types", "ModelTemplate.baseMappings"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Mapping", "1"),
                ),
            ],
        )



from __future__ import annotations


from cognite.client import data_modeling as dm

from ._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE


class PriceAreaAssetPlantsAPI(EdgeAPI):
    def list(
        self,
        from_price_area_asset: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        from_price_area_asset_space: str = DEFAULT_INSTANCE_SPACE,
        to_plant: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        to_plant_space: str = DEFAULT_INSTANCE_SPACE,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit=DEFAULT_LIMIT_READ,
    ) -> dm.EdgeList:
        """List plant edges of a price area asset.

        Args:
            from_price_area_asset: ID of the source price area asset.
            from_price_area_asset_space: Location of the price area assets.
            to_plant: ID of the target plant.
            to_plant_space: Location of the plants.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            The requested plant edges.

        Examples:

            List 5 plant edges connected to "my_price_area_asset":

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> price_area_asset = client.price_area_asset.plants_edge.list("my_price_area_asset", limit=5)

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "isPlantOf"),
            from_price_area_asset,
            from_price_area_asset_space,
            to_plant,
            to_plant_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._case import Case, CaseWrite


__all__ = [
    "PriceProdCase",
    "PriceProdCaseWrite",
    "PriceProdCaseApply",
    "PriceProdCaseList",
    "PriceProdCaseWriteList",
    "PriceProdCaseApplyList",
    "PriceProdCaseFields",
    "PriceProdCaseTextFields",
]


PriceProdCaseTextFields = Literal["price", "production"]
PriceProdCaseFields = Literal["price", "production"]

_PRICEPRODCASE_PROPERTIES_BY_FIELD = {
    "price": "price",
    "production": "production",
}


class PriceProdCase(DomainModel):
    """This represents the reading version of price prod case.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price prod case.
        data_record: The data record of the price prod case node.
        price: The price field.
        production: The production field.
        case: The case field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "PriceProdCase"
    )
    price: Union[TimeSeries, str, None] = None
    production: Union[TimeSeries, str, None] = None
    case: Union[Case, str, dm.NodeId, None] = Field(None, repr=False)

    def as_write(self) -> PriceProdCaseWrite:
        """Convert this read version of price prod case to the writing version."""
        return PriceProdCaseWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            price=self.price,
            production=self.production,
            case=self.case.as_write() if isinstance(self.case, DomainModel) else self.case,
        )

    def as_apply(self) -> PriceProdCaseWrite:
        """Convert this read version of price prod case to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceProdCaseWrite(DomainModelWrite):
    """This represents the writing version of price prod case.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price prod case.
        data_record: The data record of the price prod case node.
        price: The price field.
        production: The production field.
        case: The case field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "PriceProdCase"
    )
    price: Union[TimeSeries, str, None] = None
    production: Union[TimeSeries, str, None] = None
    case: Union[CaseWrite, str, dm.NodeId, None] = Field(None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            PriceProdCase, dm.ViewId("sp_powerops_models", "PriceProdCase", "1")
        )

        properties: dict[str, Any] = {}

        if self.price is not None or write_none:
            if isinstance(self.price, str) or self.price is None:
                properties["price"] = self.price
            else:
                properties["price"] = self.price.external_id

        if self.production is not None or write_none:
            if isinstance(self.production, str) or self.production is None:
                properties["production"] = self.production
            else:
                properties["production"] = self.production.external_id

        if self.case is not None:
            properties["case"] = {
                "space": self.space if isinstance(self.case, str) else self.case.space,
                "externalId": self.case if isinstance(self.case, str) else self.case.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.case, DomainModelWrite):
            other_resources = self.case._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.price, CogniteTimeSeries):
            resources.time_series.append(self.price)

        if isinstance(self.production, CogniteTimeSeries):
            resources.time_series.append(self.production)

        return resources


class PriceProdCaseApply(PriceProdCaseWrite):
    def __new__(cls, *args, **kwargs) -> PriceProdCaseApply:
        warnings.warn(
            "PriceProdCaseApply is deprecated and will be removed in v1.0. Use PriceProdCaseWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceProdCase.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PriceProdCaseList(DomainModelList[PriceProdCase]):
    """List of price prod cases in the read version."""

    _INSTANCE = PriceProdCase

    def as_write(self) -> PriceProdCaseWriteList:
        """Convert these read versions of price prod case to the writing versions."""
        return PriceProdCaseWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceProdCaseWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceProdCaseWriteList(DomainModelWriteList[PriceProdCaseWrite]):
    """List of price prod cases in the writing version."""

    _INSTANCE = PriceProdCaseWrite


class PriceProdCaseApplyList(PriceProdCaseWriteList): ...


def _create_price_prod_case_filter(
    view_id: dm.ViewId,
    case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if case and isinstance(case, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("case"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": case}
            )
        )
    if case and isinstance(case, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("case"), value={"space": case[0], "externalId": case[1]})
        )
    if case and isinstance(case, list) and isinstance(case[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in case],
            )
        )
    if case and isinstance(case, list) and isinstance(case[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"), values=[{"space": item[0], "externalId": item[1]} for item in case]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._mapping import Mapping, MappingWrite
    from ._model_template import ModelTemplate, ModelTemplateWrite


__all__ = [
    "ScenarioRaw",
    "ScenarioRawWrite",
    "ScenarioRawApply",
    "ScenarioRawList",
    "ScenarioRawWriteList",
    "ScenarioRawApplyList",
    "ScenarioRawFields",
    "ScenarioRawTextFields",
]


ScenarioRawTextFields = Literal[
    "name",
    "shop_version",
    "model_file",
    "commands",
    "extra_files",
    "source",
    "shop_start_specification",
    "shop_end_specification",
]
ScenarioRawFields = Literal[
    "name",
    "shop_version",
    "model_file",
    "commands",
    "extra_files",
    "source",
    "shop_start_specification",
    "shop_end_specification",
    "shop_start",
    "shop_end",
    "bid_date",
    "is_ready",
]

_SCENARIORAW_PROPERTIES_BY_FIELD = {
    "name": "name",
    "shop_version": "shopVersion",
    "model_file": "modelFile",
    "commands": "commands",
    "extra_files": "extraFiles",
    "source": "source",
    "shop_start_specification": "shopStartSpecification",
    "shop_end_specification": "shopEndSpecification",
    "shop_start": "shopStart",
    "shop_end": "shopEnd",
    "bid_date": "bidDate",
    "is_ready": "isReady",
}


class ScenarioRaw(DomainModel, protected_namespaces=()):
    """This represents the reading version of scenario raw.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the scenario raw.
        data_record: The data record of the scenario raw node.
        name: The name of the scenario to run
        shop_version: The shop version to use when running the scenario
        model_template: The model template to use when running the scenario
        model_file: The final model file to use when running the scenario (after modelTemplate is processed)
        commands: The commands to run when running the scenario
        extra_files: Extra files to include when running the scenario
        source: The source of the scenario
        shop_start_specification: The shop start specification
        shop_end_specification: The shop end specification
        shop_start: The shop start time
        shop_end: The shop end time
        bid_date: The bid date of the scenario
        is_ready: Whether the scenario is ready to be run
        mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    model_template: Union[ModelTemplate, str, dm.NodeId, None] = Field(None, repr=False, alias="modelTemplate")
    model_file: Union[str, None] = Field(None, alias="modelFile")
    commands: Optional[list[str]] = None
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    source: Optional[str] = None
    shop_start_specification: Optional[str] = Field(None, alias="shopStartSpecification")
    shop_end_specification: Optional[str] = Field(None, alias="shopEndSpecification")
    shop_start: Optional[datetime.datetime] = Field(None, alias="shopStart")
    shop_end: Optional[datetime.datetime] = Field(None, alias="shopEnd")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    is_ready: Optional[bool] = Field(None, alias="isReady")
    mappings_override: Union[list[Mapping], list[str], None] = Field(default=None, repr=False, alias="mappingsOverride")

    def as_write(self) -> ScenarioRawWrite:
        """Convert this read version of scenario raw to the writing version."""
        return ScenarioRawWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            shop_version=self.shop_version,
            model_template=(
                self.model_template.as_write() if isinstance(self.model_template, DomainModel) else self.model_template
            ),
            model_file=self.model_file,
            commands=self.commands,
            extra_files=self.extra_files,
            source=self.source,
            shop_start_specification=self.shop_start_specification,
            shop_end_specification=self.shop_end_specification,
            shop_start=self.shop_start,
            shop_end=self.shop_end,
            bid_date=self.bid_date,
            is_ready=self.is_ready,
            mappings_override=[
                mappings_override.as_write() if isinstance(mappings_override, DomainModel) else mappings_override
                for mappings_override in self.mappings_override or []
            ],
        )

    def as_apply(self) -> ScenarioRawWrite:
        """Convert this read version of scenario raw to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ScenarioRawWrite(DomainModelWrite, protected_namespaces=()):
    """This represents the writing version of scenario raw.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the scenario raw.
        data_record: The data record of the scenario raw node.
        name: The name of the scenario to run
        shop_version: The shop version to use when running the scenario
        model_template: The model template to use when running the scenario
        model_file: The final model file to use when running the scenario (after modelTemplate is processed)
        commands: The commands to run when running the scenario
        extra_files: Extra files to include when running the scenario
        source: The source of the scenario
        shop_start_specification: The shop start specification
        shop_end_specification: The shop end specification
        shop_start: The shop start time
        shop_end: The shop end time
        bid_date: The bid date of the scenario
        is_ready: Whether the scenario is ready to be run
        mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    model_template: Union[ModelTemplateWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="modelTemplate")
    model_file: Union[str, None] = Field(None, alias="modelFile")
    commands: Optional[list[str]] = None
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    source: Optional[str] = None
    shop_start_specification: Optional[str] = Field(None, alias="shopStartSpecification")
    shop_end_specification: Optional[str] = Field(None, alias="shopEndSpecification")
    shop_start: Optional[datetime.datetime] = Field(None, alias="shopStart")
    shop_end: Optional[datetime.datetime] = Field(None, alias="shopEnd")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    is_ready: Optional[bool] = Field(None, alias="isReady")
    mappings_override: Union[list[MappingWrite], list[str], None] = Field(
        default=None, repr=False, alias="mappingsOverride"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(ScenarioRaw, dm.ViewId("sp_powerops_models", "ScenarioRaw", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.shop_version is not None or write_none:
            properties["shopVersion"] = self.shop_version

        if self.model_template is not None:
            properties["modelTemplate"] = {
                "space": self.space if isinstance(self.model_template, str) else self.model_template.space,
                "externalId": (
                    self.model_template if isinstance(self.model_template, str) else self.model_template.external_id
                ),
            }

        if self.model_file is not None:
            properties["modelFile"] = self.model_file

        if self.commands is not None or write_none:
            properties["commands"] = self.commands

        if self.extra_files is not None or write_none:
            properties["extraFiles"] = self.extra_files

        if self.source is not None or write_none:
            properties["source"] = self.source

        if self.shop_start_specification is not None or write_none:
            properties["shopStartSpecification"] = self.shop_start_specification

        if self.shop_end_specification is not None or write_none:
            properties["shopEndSpecification"] = self.shop_end_specification

        if self.shop_start is not None or write_none:
            properties["shopStart"] = self.shop_start.isoformat(timespec="milliseconds") if self.shop_start else None

        if self.shop_end is not None or write_none:
            properties["shopEnd"] = self.shop_end.isoformat(timespec="milliseconds") if self.shop_end else None

        if self.bid_date is not None or write_none:
            properties["bidDate"] = self.bid_date.isoformat() if self.bid_date else None

        if self.is_ready is not None or write_none:
            properties["isReady"] = self.is_ready

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "Mapping")
        for mappings_override in self.mappings_override or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=mappings_override,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.model_template, DomainModelWrite):
            other_resources = self.model_template._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ScenarioRawApply(ScenarioRawWrite):
    def __new__(cls, *args, **kwargs) -> ScenarioRawApply:
        warnings.warn(
            "ScenarioRawApply is deprecated and will be removed in v1.0. Use ScenarioRawWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ScenarioRaw.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ScenarioRawList(DomainModelList[ScenarioRaw]):
    """List of scenario raws in the read version."""

    _INSTANCE = ScenarioRaw

    def as_write(self) -> ScenarioRawWriteList:
        """Convert these read versions of scenario raw to the writing versions."""
        return ScenarioRawWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ScenarioRawWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ScenarioRawWriteList(DomainModelWriteList[ScenarioRawWrite]):
    """List of scenario raws in the writing version."""

    _INSTANCE = ScenarioRawWrite


class ScenarioRawApplyList(ScenarioRawWriteList): ...


def _create_scenario_raw_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    shop_version: str | list[str] | None = None,
    shop_version_prefix: str | None = None,
    model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    shop_start_specification: str | list[str] | None = None,
    shop_start_specification_prefix: str | None = None,
    shop_end_specification: str | list[str] | None = None,
    shop_end_specification_prefix: str | None = None,
    min_shop_start: datetime.datetime | None = None,
    max_shop_start: datetime.datetime | None = None,
    min_shop_end: datetime.datetime | None = None,
    max_shop_end: datetime.datetime | None = None,
    min_bid_date: datetime.date | None = None,
    max_bid_date: datetime.date | None = None,
    is_ready: bool | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(shop_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopVersion"), value=shop_version))
    if shop_version and isinstance(shop_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopVersion"), values=shop_version))
    if shop_version_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopVersion"), value=shop_version_prefix))
    if model_template and isinstance(model_template, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": model_template},
            )
        )
    if model_template and isinstance(model_template, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"),
                value={"space": model_template[0], "externalId": model_template[1]},
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in model_template],
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": item[0], "externalId": item[1]} for item in model_template],
            )
        )
    if isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if isinstance(shop_start_specification, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("shopStartSpecification"), value=shop_start_specification)
        )
    if shop_start_specification and isinstance(shop_start_specification, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("shopStartSpecification"), values=shop_start_specification)
        )
    if shop_start_specification_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopStartSpecification"), value=shop_start_specification_prefix)
        )
    if isinstance(shop_end_specification, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopEndSpecification"), value=shop_end_specification))
    if shop_end_specification and isinstance(shop_end_specification, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopEndSpecification"), values=shop_end_specification))
    if shop_end_specification_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopEndSpecification"), value=shop_end_specification_prefix)
        )
    if min_shop_start is not None or max_shop_start is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("shopStart"),
                gte=min_shop_start.isoformat(timespec="milliseconds") if min_shop_start else None,
                lte=max_shop_start.isoformat(timespec="milliseconds") if max_shop_start else None,
            )
        )
    if min_shop_end is not None or max_shop_end is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("shopEnd"),
                gte=min_shop_end.isoformat(timespec="milliseconds") if min_shop_end else None,
                lte=max_shop_end.isoformat(timespec="milliseconds") if max_shop_end else None,
            )
        )
    if min_bid_date is not None or max_bid_date is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("bidDate"),
                gte=min_bid_date.isoformat() if min_bid_date else None,
                lte=max_bid_date.isoformat() if max_bid_date else None,
            )
        )
    if isinstance(is_ready, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isReady"), value=is_ready))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, field_validator
from typing_extensions import TypeAlias

from cognite.powerops.resync.config._shared import ExternalId

GeneratorName: TypeAlias = str


class Generator(BaseModel):
    name: GeneratorName
    penstock: str
    startcost: float
    p_min: float = 0.0

    start_stop_cost_time_series: Optional[ExternalId] = None  # external ID of time series with values in m
    is_available_time_series: Optional[ExternalId] = None  # external ID of boolean time series

    @property
    def external_id(self) -> ExternalId:
        return f"generator_{self.name}"

    @field_validator("penstock", mode="before")
    def to_string(cls, value):
        return str(value)

    @field_validator("startcost", mode="before")
    def to_float(cls, value):
        if isinstance(value, dict) and len(value) >= 1:
            # Timeseries with one value
            _, value = next(iter(value.items()))
            return float(value)
        return value


class GeneratorTimeSeriesMapping(BaseModel):
    generator_name: GeneratorName
    start_stop_cost: Optional[ExternalId] = None
    is_available: Optional[ExternalId] = None

    @field_validator("start_stop_cost", "is_available", mode="before")
    def parset_number_to_string(cls, value):
        return str(value) if isinstance(value, (int, float)) else value



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._plant import Plant, PlantWrite


__all__ = [
    "Watercourse",
    "WatercourseWrite",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseWriteList",
    "WatercourseApplyList",
    "WatercourseFields",
    "WatercourseTextFields",
]


WatercourseTextFields = Literal["name", "display_name", "production_obligation"]
WatercourseFields = Literal["name", "display_name", "production_obligation", "penalty_limit"]

_WATERCOURSE_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "production_obligation": "productionObligation",
    "penalty_limit": "penaltyLimit",
}


class Watercourse(DomainModel):
    """This represents the reading version of watercourse.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the watercourse.
        data_record: The data record of the watercourse node.
        name: Name for the Watercourse.
        display_name: Display name for the Watercourse.
        production_obligation: The production obligation for the Watercourse.
        penalty_limit: The penalty limit for the watercourse (used by SHOP).
        plants: The plants that are connected to the Watercourse.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "Watercourse")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    production_obligation: Union[list[TimeSeries], list[str], None] = Field(None, alias="productionObligation")
    penalty_limit: Optional[float] = Field(None, alias="penaltyLimit")
    plants: Union[list[Plant], list[str], None] = Field(default=None, repr=False)

    def as_write(self) -> WatercourseWrite:
        """Convert this read version of watercourse to the writing version."""
        return WatercourseWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            production_obligation=self.production_obligation,
            penalty_limit=self.penalty_limit,
            plants=[plant.as_write() if isinstance(plant, DomainModel) else plant for plant in self.plants or []],
        )

    def as_apply(self) -> WatercourseWrite:
        """Convert this read version of watercourse to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class WatercourseWrite(DomainModelWrite):
    """This represents the writing version of watercourse.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the watercourse.
        data_record: The data record of the watercourse node.
        name: Name for the Watercourse.
        display_name: Display name for the Watercourse.
        production_obligation: The production obligation for the Watercourse.
        penalty_limit: The penalty limit for the watercourse (used by SHOP).
        plants: The plants that are connected to the Watercourse.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "Watercourse")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    production_obligation: Union[list[TimeSeries], list[str], None] = Field(None, alias="productionObligation")
    penalty_limit: Optional[float] = Field(None, alias="penaltyLimit")
    plants: Union[list[PlantWrite], list[str], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Watercourse, dm.ViewId("power-ops-assets", "Watercourse", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.production_obligation is not None or write_none:
            properties["productionObligation"] = [
                value if isinstance(value, str) else value.external_id for value in self.production_obligation or []
            ] or None

        if self.penalty_limit is not None or write_none:
            properties["penaltyLimit"] = self.penalty_limit

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "isSubAssetOf")
        for plant in self.plants or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=plant, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        if isinstance(self.production_obligation, CogniteTimeSeries):
            resources.time_series.append(self.production_obligation)

        return resources


class WatercourseApply(WatercourseWrite):
    def __new__(cls, *args, **kwargs) -> WatercourseApply:
        warnings.warn(
            "WatercourseApply is deprecated and will be removed in v1.0. Use WatercourseWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Watercourse.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class WatercourseList(DomainModelList[Watercourse]):
    """List of watercourses in the read version."""

    _INSTANCE = Watercourse

    def as_write(self) -> WatercourseWriteList:
        """Convert these read versions of watercourse to the writing versions."""
        return WatercourseWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> WatercourseWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class WatercourseWriteList(DomainModelWriteList[WatercourseWrite]):
    """List of watercourses in the writing version."""

    _INSTANCE = WatercourseWrite


class WatercourseApplyList(WatercourseWriteList): ...


def _create_watercourse_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_penalty_limit: float | None = None,
    max_penalty_limit: float | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_penalty_limit is not None or max_penalty_limit is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("penaltyLimit"), gte=min_penalty_limit, lte=max_penalty_limit)
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations


from cognite.client import data_modeling as dm

from ._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.assets.data_classes._core import DEFAULT_INSTANCE_SPACE


class PriceAreaWatercoursesAPI(EdgeAPI):
    def list(
        self,
        from_price_area: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        from_price_area_space: str = DEFAULT_INSTANCE_SPACE,
        to_watercourse: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        to_watercourse_space: str = DEFAULT_INSTANCE_SPACE,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit=DEFAULT_LIMIT_READ,
    ) -> dm.EdgeList:
        """List watercourse edges of a price area.

        Args:
            from_price_area: ID of the source price area.
            from_price_area_space: Location of the price areas.
            to_watercourse: ID of the target watercourse.
            to_watercourse_space: Location of the watercourses.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourse edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            The requested watercourse edges.

        Examples:

            List 5 watercourse edges connected to "my_price_area":

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> price_area = client.price_area.watercourses_edge.list("my_price_area", limit=5)

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("power-ops-types", "isSubAssetOf"),
            from_price_area,
            from_price_area_space,
            to_watercourse,
            to_watercourse_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    PartialPostProcessingOutput,
    PartialPostProcessingInput,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter

if TYPE_CHECKING:
    from .alert_query import AlertQueryAPI
    from .bid_matrix_query import BidMatrixQueryAPI


class PartialPostProcessingOutputQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("partial_post_processing_output"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select(
                    [dm.query.SourceSelector(self._view_by_read_class[PartialPostProcessingOutput], ["*"])]
                ),
                result_cls=PartialPostProcessingOutput,
                max_retrieve_limit=limit,
            )
        )

    def alerts(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        retrieve_input_: bool = False,
    ) -> AlertQueryAPI[T_DomainModelList]:
        """Query along the alert edges of the partial post processing output.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of alert edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            retrieve_input_: Whether to retrieve the input for each partial post processing output or not.

        Returns:
            AlertQueryAPI: The query API for the alert.
        """
        from .alert_query import AlertQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "calculationIssue"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("alerts"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        if retrieve_input_:
            self._query_append_input_(from_)
        return AlertQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def partial_matrices(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        retrieve_input_: bool = False,
    ) -> BidMatrixQueryAPI[T_DomainModelList]:
        """Query along the partial matrice edges of the partial post processing output.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial matrice edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            retrieve_input_: Whether to retrieve the input for each partial post processing output or not.

        Returns:
            BidMatrixQueryAPI: The query API for the bid matrix.
        """
        from .bid_matrix_query import BidMatrixQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "BidMatrix"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("partial_matrices"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        if retrieve_input_:
            self._query_append_input_(from_)
        return BidMatrixQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def query(
        self,
        retrieve_input_: bool = False,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Args:
            retrieve_input_: Whether to retrieve the input for each partial post processing output or not.

        Returns:
            The list of the source nodes of the query.

        """
        from_ = self._builder[-1].name
        if retrieve_input_:
            self._query_append_input_(from_)
        return self._query()

    def _query_append_input_(self, from_: str) -> None:
        view_id = self._view_by_read_class[PartialPostProcessingInput]
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("input_"),
                expression=dm.query.NodeResultSetExpression(
                    filter=dm.filters.HasData(views=[view_id]),
                    from_=from_,
                    through=self._view_by_read_class[PartialPostProcessingOutput].as_property_ref("input"),
                    direction="outwards",
                ),
                select=dm.query.Select([dm.query.SourceSelector(view_id, ["*"])]),
                max_retrieve_limit=-1,
                result_cls=PartialPostProcessingInput,
            ),
        )



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._market_configuration import MarketConfiguration, MarketConfigurationWrite


__all__ = [
    "BidConfiguration",
    "BidConfigurationWrite",
    "BidConfigurationApply",
    "BidConfigurationList",
    "BidConfigurationWriteList",
    "BidConfigurationApplyList",
]


class BidConfiguration(DomainModel):
    """This represents the reading version of bid configuration.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration.
        data_record: The data record of the bid configuration node.
        market_configuration: The bid method related to the bid configuration
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    market_configuration: Union[MarketConfiguration, str, dm.NodeId, None] = Field(
        None, repr=False, alias="marketConfiguration"
    )

    def as_write(self) -> BidConfigurationWrite:
        """Convert this read version of bid configuration to the writing version."""
        return BidConfigurationWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            market_configuration=(
                self.market_configuration.as_write()
                if isinstance(self.market_configuration, DomainModel)
                else self.market_configuration
            ),
        )

    def as_apply(self) -> BidConfigurationWrite:
        """Convert this read version of bid configuration to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationWrite(DomainModelWrite):
    """This represents the writing version of bid configuration.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration.
        data_record: The data record of the bid configuration node.
        market_configuration: The bid method related to the bid configuration
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    market_configuration: Union[MarketConfigurationWrite, str, dm.NodeId, None] = Field(
        None, repr=False, alias="marketConfiguration"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BidConfiguration, dm.ViewId("sp_powerops_models", "BidConfiguration", "1")
        )

        properties: dict[str, Any] = {}

        if self.market_configuration is not None:
            properties["marketConfiguration"] = {
                "space": self.space if isinstance(self.market_configuration, str) else self.market_configuration.space,
                "externalId": (
                    self.market_configuration
                    if isinstance(self.market_configuration, str)
                    else self.market_configuration.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.market_configuration, DomainModelWrite):
            other_resources = self.market_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidConfigurationApply(BidConfigurationWrite):
    def __new__(cls, *args, **kwargs) -> BidConfigurationApply:
        warnings.warn(
            "BidConfigurationApply is deprecated and will be removed in v1.0. Use BidConfigurationWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidConfiguration.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidConfigurationList(DomainModelList[BidConfiguration]):
    """List of bid configurations in the read version."""

    _INSTANCE = BidConfiguration

    def as_write(self) -> BidConfigurationWriteList:
        """Convert these read versions of bid configuration to the writing versions."""
        return BidConfigurationWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidConfigurationWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationWriteList(DomainModelWriteList[BidConfigurationWrite]):
    """List of bid configurations in the writing version."""

    _INSTANCE = BidConfigurationWrite


class BidConfigurationApplyList(BidConfigurationWriteList): ...


def _create_bid_configuration_filter(
    view_id: dm.ViewId,
    market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if market_configuration and isinstance(market_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": market_configuration},
            )
        )
    if market_configuration and isinstance(market_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": market_configuration[0], "externalId": market_configuration[1]},
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in market_configuration],
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in market_configuration],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from pathlib import Path

from rich import print

from cognite.powerops import PowerOpsClient
from cognite.powerops.resync.v2.shop_to_assets import PowerAssetImporter


def apply2(config_dir: Path, client: PowerOpsClient | None = None) -> None:
    client = client or PowerOpsClient.from_settings()

    importer = PowerAssetImporter.from_directory(config_dir / "production")

    assets = importer.to_power_assets()

    client.v1.upsert(assets)

    print(f"Upserted {len(assets)} assets")



from datetime import datetime

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
from matplotlib.axes import Axes


def create_time_series_plot() -> Axes:
    fig, ax = plt.subplots(figsize=(10, 10))
    fig.autofmt_xdate()
    return ax


def ax_plot_time_series(
    ax: Axes,
    time_series: dict[datetime, float],
    label: str,
):
    ax.plot(
        time_series.keys(),
        time_series.values(),
        linestyle="-",
        marker=".",
        label=label,
        alpha=0.7,
    )
    ax.xaxis.set_major_formatter(
        mdates.DateFormatter("%d. %b %y %H:%M"),
    )



from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)


__all__ = [
    "BidMethod",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
]


BidMethodTextFields = Literal["name"]
BidMethodFields = Literal["name"]

_BIDMETHOD_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class BidMethod(DomainModel):
    """This represents the reading version of bid method.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method.
        data_record: The data record of the bid method node.
        name: Name for the BidMethod
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str

    def as_write(self) -> BidMethodWrite:
        """Convert this read version of bid method to the writing version."""
        return BidMethodWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
        )

    def as_apply(self) -> BidMethodWrite:
        """Convert this read version of bid method to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMethodWrite(DomainModelWrite):
    """This represents the writing version of bid method.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method.
        data_record: The data record of the bid method node.
        name: Name for the BidMethod
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(BidMethod, dm.ViewId("power-ops-shared", "BidMethod", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class BidMethodApply(BidMethodWrite):
    def __new__(cls, *args, **kwargs) -> BidMethodApply:
        warnings.warn(
            "BidMethodApply is deprecated and will be removed in v1.0. Use BidMethodWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidMethod.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidMethodList(DomainModelList[BidMethod]):
    """List of bid methods in the read version."""

    _INSTANCE = BidMethod

    def as_write(self) -> BidMethodWriteList:
        """Convert these read versions of bid method to the writing versions."""
        return BidMethodWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidMethodWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMethodWriteList(DomainModelWriteList[BidMethodWrite]):
    """List of bid methods in the writing version."""

    _INSTANCE = BidMethodWrite


class BidMethodApplyList(BidMethodWriteList): ...


def _create_bid_method_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    TaskDispatcherShopInput,
    BidConfigurationShop,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter


class TaskDispatcherShopInputQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("task_dispatcher_shop_input"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select(
                    [dm.query.SourceSelector(self._view_by_read_class[TaskDispatcherShopInput], ["*"])]
                ),
                result_cls=TaskDispatcherShopInput,
                max_retrieve_limit=limit,
            )
        )

    def query(
        self,
        retrieve_bid_configuration: bool = False,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Args:
            retrieve_bid_configuration: Whether to retrieve the bid configuration for each task dispatcher shop input or not.

        Returns:
            The list of the source nodes of the query.

        """
        from_ = self._builder[-1].name
        if retrieve_bid_configuration:
            self._query_append_bid_configuration(from_)
        return self._query()

    def _query_append_bid_configuration(self, from_: str) -> None:
        view_id = self._view_by_read_class[BidConfigurationShop]
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("bid_configuration"),
                expression=dm.query.NodeResultSetExpression(
                    filter=dm.filters.HasData(views=[view_id]),
                    from_=from_,
                    through=self._view_by_read_class[TaskDispatcherShopInput].as_property_ref("bidConfiguration"),
                    direction="outwards",
                ),
                select=dm.query.Select([dm.query.SourceSelector(view_id, ["*"])]),
                max_retrieve_limit=-1,
                result_cls=BidConfigurationShop,
            ),
        )



import datetime
import secrets
from typing import Optional


def external_id_to_name(external_id: str) -> str:
    return external_id.replace("/", "_").replace(":", "-")


def unique_short_str(nbytes: int) -> str:
    return secrets.token_hex(nbytes=nbytes)


def new_external_id(prefix: str = "POWEROPS_SHOP_RUN", now: Optional[datetime.datetime] = None) -> str:
    if now is None:
        now = datetime.datetime.now(datetime.timezone.utc)
    now_isoformat = now.isoformat().replace("+00:00", "Z")
    return f"{prefix}_{now_isoformat}_{unique_short_str(3)}"



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    PartialPostProcessingInput,
    MarketConfiguration,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter

if TYPE_CHECKING:
    from .bid_matrix_raw_query import BidMatrixRawQueryAPI


class PartialPostProcessingInputQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("partial_post_processing_input"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select(
                    [dm.query.SourceSelector(self._view_by_read_class[PartialPostProcessingInput], ["*"])]
                ),
                result_cls=PartialPostProcessingInput,
                max_retrieve_limit=limit,
            )
        )

    def partial_bid_matrices_raw(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        retrieve_market_config: bool = False,
    ) -> BidMatrixRawQueryAPI[T_DomainModelList]:
        """Query along the partial bid matrices raw edges of the partial post processing input.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrices raw edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            retrieve_market_config: Whether to retrieve the market config for each partial post processing input or not.

        Returns:
            BidMatrixRawQueryAPI: The query API for the bid matrix raw.
        """
        from .bid_matrix_raw_query import BidMatrixRawQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "partialBidMatricesRaw"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("partial_bid_matrices_raw"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        if retrieve_market_config:
            self._query_append_market_config(from_)
        return BidMatrixRawQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def query(
        self,
        retrieve_market_config: bool = False,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Args:
            retrieve_market_config: Whether to retrieve the market config for each partial post processing input or not.

        Returns:
            The list of the source nodes of the query.

        """
        from_ = self._builder[-1].name
        if retrieve_market_config:
            self._query_append_market_config(from_)
        return self._query()

    def _query_append_market_config(self, from_: str) -> None:
        view_id = self._view_by_read_class[MarketConfiguration]
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("market_config"),
                expression=dm.query.NodeResultSetExpression(
                    filter=dm.filters.HasData(views=[view_id]),
                    from_=from_,
                    through=self._view_by_read_class[PartialPostProcessingInput].as_property_ref("marketConfig"),
                    direction="outwards",
                ),
                select=dm.query.Select([dm.query.SourceSelector(view_id, ["*"])]),
                max_retrieve_limit=-1,
                result_cls=MarketConfiguration,
            ),
        )



from cognite.client import CogniteClient

from cognite.powerops.client.data_set_api import DataSetsAPI
from cognite.powerops.client.shop.api.shop_run_api import ShopRunsAPI

from .api.shop_result_files_api import ShopFilesAPI
from .api.shop_results_api import ShopRunResultsAPI


class SHOPRunAPI:
    """
    The ShopClient is a client for the Shop API. It is built on top of the Cognite Python SDK

    Examples:

    Initialization of SHOP client:

        >>> from cognite.powerops.client import PowerOpsClient
        >>> client = PowerOpsClient()
        >>> shop_client = client.shop

    """

    def __init__(self, client: CogniteClient, data_set_api: DataSetsAPI, cogshop_version: str):
        self.files = ShopFilesAPI(client)
        self.results = ShopRunResultsAPI(client, self.files)
        self.runs = ShopRunsAPI(client, data_set_api, self.results, cogshop_version)



from __future__ import annotations

import json
from datetime import datetime
from pathlib import Path
from typing import Optional

import yaml
from cognite.client.data_classes import FileMetadata
from pydantic import Field, field_validator
from typing_extensions import TypeAlias

from cognite.powerops.client._generated.cogshop1 import data_classes as cogshop_v1
from cognite.powerops.prerun_transformations.transformations import Transformation as TransformationV2
from cognite.powerops.resync import config
from cognite.powerops.resync.config import Transformation
from cognite.powerops.resync.models.base import CDFFile, Model
from cognite.powerops.utils.serialization import load_yaml

ExternalID: TypeAlias = str


class CogShopCore(Model):
    shop_files: list[CDFFile] = Field(default_factory=list)

    @field_validator("shop_files", mode="after")
    def ordering(cls, value: list[CDFFile]) -> list[CDFFile]:
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.shop_files = self.ordering(self.shop_files)


def _to_shop_model_file(
    watercourse_name, model_file: Path, processed_model_file: Path, write_to_local: Optional[bool] = True
) -> CDFFile:
    processed_model = _to_model_without_timeseries(
        yaml_raw_path=str(model_file),
        # Todo Move this hardcoded configuration to a config file
        non_time_series_attributes_to_remove={"reservoir.start_vol", "generator.initial_state"},
        encoding="utf-8",
    )
    yaml_content = yaml.safe_dump(processed_model, allow_unicode=True, sort_keys=False)
    file_content = yaml_content.encode("utf-8")
    if write_to_local:
        dumped = yaml.dump(processed_model, allow_unicode=True, sort_keys=False)
        processed_model_file.write_text(dumped, encoding="utf-8")
    external_id = f"SHOP_{watercourse_name}_{processed_model_file.stem}"
    return _create_shop_file(
        file_content,
        external_id,
        {"shop:type": "case", "shop:watercourse": watercourse_name, "shop:file_name": processed_model_file.stem},
    )


def _to_shop_files(watercourses_shop: list[config.ShopFileConfig]) -> list[CDFFile]:
    shop_files = []
    for shop_config in watercourses_shop:
        file_content = Path(shop_config.file_path).read_bytes()
        shop_file = _create_shop_file(
            file_content,
            shop_config.external_id,
            {
                "shop:type": shop_config.cogshop_file_type,
                "shop:watercourse": shop_config.watercourse_name,
                "shop:file_name": shop_config.file_path.stem,
            },
        )
        shop_files.append(shop_file)
    return shop_files


def _create_shop_file(file_content: bytes, external_id: str, metadata: dict[str, str]) -> CDFFile:
    return CDFFile(
        meta=FileMetadata(
            external_id=external_id,
            name=external_id,
            metadata=metadata,
            source="PowerOps bootstrap",
            mime_type="text/plain",
        ),
        content=file_content,
    )


def _is_time_series(shop_attribute_value) -> bool:
    return isinstance(shop_attribute_value, dict) and isinstance(next(iter(shop_attribute_value)), datetime)


def _to_model_without_timeseries(
    yaml_raw_path: str, non_time_series_attributes_to_remove: set[str] | None = None, encoding=None
) -> dict:
    non_time_series_attributes_to_remove = non_time_series_attributes_to_remove or set()

    model_incl_time_and_connections = load_yaml(Path(yaml_raw_path), encoding=encoding, clean_data=True)

    model = model_incl_time_and_connections["model"]

    for object_type, objects in model.items():
        for object_attributes in objects.values():
            for attribute_name, attribute_data in list(
                object_attributes.items()
            ):  # Needs list() since we pop while iterating over the dict
                if f"{object_type}.{attribute_name}" in non_time_series_attributes_to_remove:
                    object_attributes.pop(attribute_name)

                elif _is_time_series(attribute_data):
                    object_attributes.pop(attribute_name)

    model_dict = {"model": model}

    # TODO: remove this when standardizing parameters SHOP file structure (model vs model + connections)
    if "connections" in model_incl_time_and_connections:
        model_dict["connections"] = model_incl_time_and_connections["connections"]

    return model_dict


def _create_transformation(order: int, transformation: dict | config.Transformation) -> cogshop_v1.TransformationApply:
    if isinstance(transformation, dict):
        transformation = config.Transformation(**transformation)
    dumped_kwargs = json.dumps(transformation.kwargs or {}, separators=(",", ":"))
    external_id = f"Tr_{transformation.transformation.name}_{dumped_kwargs}_{order}"
    return cogshop_v1.TransformationApply(
        external_id=external_id, method=transformation.transformation.name, arguments=dumped_kwargs, order=order
    )


def _create_transformationV2(
    order: int, transformation: dict | TransformationV2 | Transformation
) -> cogshop_v1.TransformationApply:
    """
    Adapter betweeen transformationsV2 pydantic instances to CogShop model 1 FDM instances
    """
    if isinstance(transformation, dict):
        transformation = TransformationV2.load(transformation)
    # TODO: temp hack to be able to read in old transformations from price scenarios in Dayahead and Rkom
    # TODO: remove once market configs can load the new price scenarios config
    elif isinstance(transformation, Transformation):
        transformation = transformations_v2_transformer(transformation)

    dumped_parameters_args = json.dumps(
        transformation.parameters_to_dict() or {}, separators=(",", ":"), ensure_ascii=False
    )
    external_id = f"Tr2_{transformation.name}_{dumped_parameters_args}_{order}"
    if len(external_id) > 255:
        external_id = f"Tr2_{transformation.name}_{dumped_parameters_args[:len(dumped_parameters_args) // 2]}_{order}"
    return cogshop_v1.TransformationApply(
        external_id=external_id, method=transformation.name, arguments=dumped_parameters_args, order=order
    )


def transformations_v2_transformer(
    transformation: Transformation | dict, object_type: Optional[str] = None, object_name: Optional[str] = None
) -> TransformationV2:
    """
    Adapter that converts old transformation to an instance of new transformationsV2.
    From these instances, the transformationsV2 to FDM adapter "_create_transformationV2"
    can be called to create DM nodes and edges
    """
    if isinstance(transformation, dict):
        transformation = config.Transformation(**transformation)

    transformation_dict = {}
    if transformation.transformation.name == "ADD":
        transformation_dict = {"AddConstant": {"parameters": {"constant": transformation.kwargs.get("value")}}}
    elif transformation.transformation.name in ["ADD_FROM_OFFSET", "DYNAMIC_ADD_FROM_OFFSET"]:
        transformation_dict = {
            "AddFromOffset": {
                "parameters": {
                    "relative_datapoints": [
                        {"offset_minute": k, "offset_value": v} for k, v in transformation.kwargs.items()
                    ]
                }
            }
        }
    elif transformation.transformation.name == "ADD_WATER_IN_TRANSIT":
        gate_or_plant = "gate" if "gate_name" in transformation.kwargs.keys() else "plant"
        transformation_dict = {
            "AddWaterInTransit": {
                "parameters": {
                    "discharge_ts_external_id": transformation.kwargs.get("external_id"),
                    "transit_object_type": gate_or_plant,
                    "transit_object_name": transformation.kwargs.get(f"{gate_or_plant}_name"),
                }
            }
        }
    elif transformation.transformation.name == "ADD_WATER_IN_TRANSIT":
        gate_or_plant = "gate" if "gate_name" in transformation.kwargs.keys() else "plant"
        transformation_dict = {
            "AddWaterInTransit": {
                "parameters": {
                    "discharge_ts_external_id": transformation.kwargs.get("external_id"),
                    "transit_object_type": gate_or_plant,
                    "transit_object_name": transformation.kwargs.get(f"{gate_or_plant}_name"),
                }
            }
        }

    elif transformation.transformation.name == "MULTIPLY":
        transformation_dict = {"MultiplyConstant": {"parameters": {"constant": transformation.kwargs.get("value")}}}
    elif transformation.transformation.name == "MULTIPLY_FROM_OFFSET":
        transformation_dict = {
            "MultiplyFromOffset": {
                "parameters": {
                    "relative_datapoints": [
                        {"offset_minute": k, "offset_value": v} for k, v in transformation.kwargs.items()
                    ]
                }
            }
        }
    elif transformation.transformation.name == "TO_BOOL":
        transformation_dict = {"ToBool": None}
    elif transformation.transformation.name == "TO_INT":
        transformation_dict = {"ToInt": None}
    elif transformation.transformation.name in [
        "ZERO_IF_NOT_ONE",
        "GATE_SCHEDULE_FLAG_VALUE_MAPPING",
        "GENERATOR_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING",
        "PLANT_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING",
    ]:
        transformation_dict = {"ZeroIfNotOne": None}
    elif transformation.transformation.name == "ONE_IF_TWO":
        transformation_dict = {"OneIfTwo": None}
    elif transformation.transformation.name in ["STATIC", "DYNAMIC_STATIC"]:
        transformation_dict = {
            "StaticValues": {
                "parameters": {
                    "relative_datapoints": [
                        {"offset_minute": k, "offset_value": v} for k, v in transformation.kwargs.items()
                    ]
                }
            }
        }
    elif transformation.transformation.name == "RESERVOIR_LEVEL_TO_VOLUME":
        transformation_dict = {
            "HeightToVolume": {"parameters": {"object_type": object_type, "object_name": object_name}}
        }
    elif transformation.transformation.name in ["DO_NOTHING", "GATE_OPENING_METER_TO_PERCENT"]:
        transformation_dict = {"DoNothing": None}
    else:
        raise NotImplementedError(f"Unknown transformation {transformation.transformation.name}")

    return TransformationV2.load(transformation_dict)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    BidCalculationTask,
    BidCalculationTaskWrite,
    BidCalculationTaskFields,
    BidCalculationTaskList,
    BidCalculationTaskWriteList,
)
from cognite.powerops.client._generated.v1.data_classes._bid_calculation_task import (
    _BIDCALCULATIONTASK_PROPERTIES_BY_FIELD,
    _create_bid_calculation_task_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .bid_calculation_task_query import BidCalculationTaskQueryAPI


class BidCalculationTaskAPI(NodeAPI[BidCalculationTask, BidCalculationTaskWrite, BidCalculationTaskList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[BidCalculationTask]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidCalculationTask,
            class_list=BidCalculationTaskList,
            class_write_list=BidCalculationTaskWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id

    def __call__(
        self,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> BidCalculationTaskQueryAPI[BidCalculationTaskList]:
        """Query starting at bid calculation tasks.

        Args:
            plant: The plant to filter on.
            min_bid_date: The minimum value of the bid date to filter on.
            max_bid_date: The maximum value of the bid date to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid calculation tasks to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for bid calculation tasks.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_bid_calculation_task_filter(
            self._view_id,
            plant,
            min_bid_date,
            max_bid_date,
            price_area,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(BidCalculationTaskList)
        return BidCalculationTaskQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        bid_calculation_task: BidCalculationTaskWrite | Sequence[BidCalculationTaskWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) bid calculation tasks.

        Args:
            bid_calculation_task: Bid calculation task or sequence of bid calculation tasks to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new bid_calculation_task:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import BidCalculationTaskWrite
                >>> client = PowerOpsModelsV1Client()
                >>> bid_calculation_task = BidCalculationTaskWrite(external_id="my_bid_calculation_task", ...)
                >>> result = client.bid_calculation_task.apply(bid_calculation_task)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.bid_calculation_task.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(bid_calculation_task, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more bid calculation task.

        Args:
            external_id: External id of the bid calculation task to delete.
            space: The space where all the bid calculation task are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete bid_calculation_task by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.bid_calculation_task.delete("my_bid_calculation_task")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.bid_calculation_task.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> BidCalculationTask | None: ...

    @overload
    def retrieve(
        self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> BidCalculationTaskList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> BidCalculationTask | BidCalculationTaskList | None:
        """Retrieve one or more bid calculation tasks by id(s).

        Args:
            external_id: External id or list of external ids of the bid calculation tasks.
            space: The space where all the bid calculation tasks are located.

        Returns:
            The requested bid calculation tasks.

        Examples:

            Retrieve bid_calculation_task by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> bid_calculation_task = client.bid_calculation_task.retrieve("my_bid_calculation_task")

        """
        return self._retrieve(external_id, space)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: BidCalculationTaskFields | Sequence[BidCalculationTaskFields] | None = None,
        group_by: None = None,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: BidCalculationTaskFields | Sequence[BidCalculationTaskFields] | None = None,
        group_by: BidCalculationTaskFields | Sequence[BidCalculationTaskFields] = None,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: BidCalculationTaskFields | Sequence[BidCalculationTaskFields] | None = None,
        group_by: BidCalculationTaskFields | Sequence[BidCalculationTaskFields] | None = None,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across bid calculation tasks

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            plant: The plant to filter on.
            min_bid_date: The minimum value of the bid date to filter on.
            max_bid_date: The maximum value of the bid date to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid calculation tasks to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count bid calculation tasks in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.bid_calculation_task.aggregate("count", space="my_space")

        """

        filter_ = _create_bid_calculation_task_filter(
            self._view_id,
            plant,
            min_bid_date,
            max_bid_date,
            price_area,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _BIDCALCULATIONTASK_PROPERTIES_BY_FIELD,
            property,
            group_by,
            None,
            None,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: BidCalculationTaskFields,
        interval: float,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for bid calculation tasks

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            plant: The plant to filter on.
            min_bid_date: The minimum value of the bid date to filter on.
            max_bid_date: The maximum value of the bid date to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid calculation tasks to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_bid_calculation_task_filter(
            self._view_id,
            plant,
            min_bid_date,
            max_bid_date,
            price_area,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _BIDCALCULATIONTASK_PROPERTIES_BY_FIELD,
            None,
            None,
            limit,
            filter_,
        )

    def list(
        self,
        plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_bid_date: datetime.date | None = None,
        max_bid_date: datetime.date | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidCalculationTaskList:
        """List/filter bid calculation tasks

        Args:
            plant: The plant to filter on.
            min_bid_date: The minimum value of the bid date to filter on.
            max_bid_date: The maximum value of the bid date to filter on.
            price_area: The price area to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid calculation tasks to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            List of requested bid calculation tasks

        Examples:

            List bid calculation tasks and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> bid_calculation_tasks = client.bid_calculation_task.list(limit=5)

        """
        filter_ = _create_bid_calculation_task_filter(
            self._view_id,
            plant,
            min_bid_date,
            max_bid_date,
            price_area,
            external_id_prefix,
            space,
            filter,
        )
        return self._list(limit=limit, filter=filter_)



import re

from cognite.client import CogniteClient

from cognite.powerops.resync.config._shared import TimeSeriesMapping


def filter_time_series_mappings(mapping: TimeSeriesMapping, client: CogniteClient) -> TimeSeriesMapping:
    """Filters out mapping entries where the expected CDF time series is not found."""
    if not mapping.rows:
        print_warning("TimeSeriesMapping is empty! No filtering possible!")
        return mapping

    external_ids_in_mapping = {
        mapping.time_series_external_id for mapping in mapping if mapping.time_series_external_id is not None
    }

    external_ids_in_cdf = {
        time_series.external_id
        for time_series in client.time_series.retrieve_multiple(
            external_ids=list(external_ids_in_mapping), ignore_unknown_ids=True
        )
    }

    if missing_from_cdf := external_ids_in_mapping - external_ids_in_cdf:
        print_warning(f"Time series found in mapping but missing from CDF: {missing_from_cdf}")

    return TimeSeriesMapping(
        rows=[
            entry
            for entry in mapping
            if entry.time_series_external_id in external_ids_in_cdf or entry.time_series_external_id is None
        ]
    )


def merge_and_keep_last_mapping_if_overlap(*mappings: TimeSeriesMapping) -> TimeSeriesMapping:
    """Within each TimeSeriesMapping keep the last TimeSeriesMappingEntry in cases
    where they have the same `shop_model_path`. Across mappings keep the last TimeSeriesMappingEntry
    from the last TimeSeriesMapping.
    """

    keep_last_across_lists = {entry.shop_model_path: entry for mapping in mappings for entry in mapping}

    return TimeSeriesMapping(rows=list(keep_last_across_lists.values()))


def special_case_handle_gate_number(name: str) -> None:
    """Must handle any gate numbers above 1 if found in other sources than YAML"""
    # TODO: extend to handle special case if needed
    if re.search(pattern=r"L[2-9]", string=name):
        print_warning(f"Potential gate {name} not in YAML!")


def print_warning(s: str) -> None:
    """Adds some nice colors to the printed text :)"""
    print(f"\033[91m[WARNING] {s}\033[0m")



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)
from ._bid_configuration import BidConfiguration, BidConfigurationWrite

if TYPE_CHECKING:
    from ._bid_method_shop_multi_scenario import BidMethodSHOPMultiScenario, BidMethodSHOPMultiScenarioWrite
    from ._market_configuration import MarketConfiguration, MarketConfigurationWrite
    from ._plant_shop import PlantShop, PlantShopWrite
    from ._price_area import PriceArea, PriceAreaWrite
    from ._watercourse_shop import WatercourseShop, WatercourseShopWrite


__all__ = [
    "BidConfigurationShop",
    "BidConfigurationShopWrite",
    "BidConfigurationShopApply",
    "BidConfigurationShopList",
    "BidConfigurationShopWriteList",
    "BidConfigurationShopApplyList",
    "BidConfigurationShopFields",
    "BidConfigurationShopTextFields",
]


BidConfigurationShopTextFields = Literal["name"]
BidConfigurationShopFields = Literal["name"]

_BIDCONFIGURATIONSHOP_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class BidConfigurationShop(BidConfiguration):
    """This represents the reading version of bid configuration shop.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration shop.
        data_record: The data record of the bid configuration shop node.
        market_configuration: The bid method related to the bid configuration
        name: The name of the bid configuration
        method: The bid method related to the bid configuration
        price_area: The price area related to the bid configuration
        plants_shop: The plants modelled in the shop runs
        watercourses_shop: The watercourses modelled in the shop runs
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "BidConfigurationShop"
    )
    name: Optional[str] = None
    method: Union[BidMethodSHOPMultiScenario, str, dm.NodeId, None] = Field(None, repr=False)
    price_area: Union[PriceArea, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    plants_shop: Union[list[PlantShop], list[str], None] = Field(default=None, repr=False, alias="plantsShop")
    watercourses_shop: Union[list[WatercourseShop], list[str], None] = Field(
        default=None, repr=False, alias="watercoursesShop"
    )

    def as_write(self) -> BidConfigurationShopWrite:
        """Convert this read version of bid configuration shop to the writing version."""
        return BidConfigurationShopWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            market_configuration=(
                self.market_configuration.as_write()
                if isinstance(self.market_configuration, DomainModel)
                else self.market_configuration
            ),
            name=self.name,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            price_area=self.price_area.as_write() if isinstance(self.price_area, DomainModel) else self.price_area,
            plants_shop=[
                plants_shop.as_write() if isinstance(plants_shop, DomainModel) else plants_shop
                for plants_shop in self.plants_shop or []
            ],
            watercourses_shop=[
                watercourses_shop.as_write() if isinstance(watercourses_shop, DomainModel) else watercourses_shop
                for watercourses_shop in self.watercourses_shop or []
            ],
        )

    def as_apply(self) -> BidConfigurationShopWrite:
        """Convert this read version of bid configuration shop to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationShopWrite(BidConfigurationWrite):
    """This represents the writing version of bid configuration shop.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration shop.
        data_record: The data record of the bid configuration shop node.
        market_configuration: The bid method related to the bid configuration
        name: The name of the bid configuration
        method: The bid method related to the bid configuration
        price_area: The price area related to the bid configuration
        plants_shop: The plants modelled in the shop runs
        watercourses_shop: The watercourses modelled in the shop runs
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "BidConfigurationShop"
    )
    name: Optional[str] = None
    method: Union[BidMethodSHOPMultiScenarioWrite, str, dm.NodeId, None] = Field(None, repr=False)
    price_area: Union[PriceAreaWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    plants_shop: Union[list[PlantShopWrite], list[str], None] = Field(default=None, repr=False, alias="plantsShop")
    watercourses_shop: Union[list[WatercourseShopWrite], list[str], None] = Field(
        default=None, repr=False, alias="watercoursesShop"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BidConfigurationShop, dm.ViewId("sp_powerops_models", "BidConfigurationShop", "1")
        )

        properties: dict[str, Any] = {}

        if self.market_configuration is not None:
            properties["marketConfiguration"] = {
                "space": self.space if isinstance(self.market_configuration, str) else self.market_configuration.space,
                "externalId": (
                    self.market_configuration
                    if isinstance(self.market_configuration, str)
                    else self.market_configuration.external_id
                ),
            }

        if self.name is not None or write_none:
            properties["name"] = self.name

        if self.method is not None:
            properties["method"] = {
                "space": self.space if isinstance(self.method, str) else self.method.space,
                "externalId": self.method if isinstance(self.method, str) else self.method.external_id,
            }

        if self.price_area is not None:
            properties["priceArea"] = {
                "space": self.space if isinstance(self.price_area, str) else self.price_area.space,
                "externalId": self.price_area if isinstance(self.price_area, str) else self.price_area.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.plantsShop")
        for plants_shop in self.plants_shop or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=plants_shop, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("sp_powerops_types", "BidConfiguration.watercoursesShop")
        for watercourses_shop in self.watercourses_shop or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=watercourses_shop,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.market_configuration, DomainModelWrite):
            other_resources = self.market_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.method, DomainModelWrite):
            other_resources = self.method._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.price_area, DomainModelWrite):
            other_resources = self.price_area._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidConfigurationShopApply(BidConfigurationShopWrite):
    def __new__(cls, *args, **kwargs) -> BidConfigurationShopApply:
        warnings.warn(
            "BidConfigurationShopApply is deprecated and will be removed in v1.0. Use BidConfigurationShopWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidConfigurationShop.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidConfigurationShopList(DomainModelList[BidConfigurationShop]):
    """List of bid configuration shops in the read version."""

    _INSTANCE = BidConfigurationShop

    def as_write(self) -> BidConfigurationShopWriteList:
        """Convert these read versions of bid configuration shop to the writing versions."""
        return BidConfigurationShopWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidConfigurationShopWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationShopWriteList(DomainModelWriteList[BidConfigurationShopWrite]):
    """List of bid configuration shops in the writing version."""

    _INSTANCE = BidConfigurationShopWrite


class BidConfigurationShopApplyList(BidConfigurationShopWriteList): ...


def _create_bid_configuration_shop_filter(
    view_id: dm.ViewId,
    market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if market_configuration and isinstance(market_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": market_configuration},
            )
        )
    if market_configuration and isinstance(market_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": market_configuration[0], "externalId": market_configuration[1]},
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in market_configuration],
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in market_configuration],
            )
        )
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if method and isinstance(method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("method"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": method}
            )
        )
    if method and isinstance(method, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("method"), value={"space": method[0], "externalId": method[1]})
        )
    if method and isinstance(method, list) and isinstance(method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in method],
            )
        )
    if method and isinstance(method, list) and isinstance(method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"), values=[{"space": item[0], "externalId": item[1]} for item in method]
            )
        )
    if price_area and isinstance(price_area, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_area}
            )
        )
    if price_area and isinstance(price_area, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": price_area[0], "externalId": price_area[1]}
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_area],
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_area],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from ._core import (
    DataRecord,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    ResourcesWriteResult,
)
from ._alert import (
    Alert,
    AlertApply,
    AlertApplyList,
    AlertFields,
    AlertList,
    AlertTextFields,
    AlertWrite,
    AlertWriteList,
)
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentList,
    BidDocumentTextFields,
    BidDocumentWrite,
    BidDocumentWriteList,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodList,
    BidMethodTextFields,
    BidMethodWrite,
    BidMethodWriteList,
)
from ._bid_row import (
    BidRow,
    BidRowApply,
    BidRowApplyList,
    BidRowFields,
    BidRowList,
    BidRowTextFields,
    BidRowWrite,
    BidRowWriteList,
)
from ._price_area import PriceArea, PriceAreaFields, PriceAreaList, PriceAreaTextFields

BidDocument.model_rebuild()
BidDocumentWrite.model_rebuild()
BidDocumentApply.model_rebuild()
BidRow.model_rebuild()
BidRowWrite.model_rebuild()
BidRowApply.model_rebuild()

__all__ = [
    "DataRecord",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "ResourcesWriteResult",
    "Alert",
    "AlertWrite",
    "AlertApply",
    "AlertList",
    "AlertWriteList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BidDocument",
    "BidDocumentWrite",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidMethod",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "BidRow",
    "BidRowWrite",
    "BidRowApply",
    "BidRowList",
    "BidRowWriteList",
    "BidRowApplyList",
    "BidRowFields",
    "BidRowTextFields",
    "PriceArea",
    "PriceAreaList",
    "PriceAreaFields",
    "PriceAreaTextFields",
]



from __future__ import annotations

from itertools import product
from typing import ClassVar

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import ContainerId
from pydantic import Field, field_validator

from cognite.powerops.client import PowerOpsClient
from cognite.powerops.client._generated.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1
from cognite.powerops.client.data_classes import cogshop1 as cogshop_v1
from cognite.powerops.resync.models._shared_v1_v2.cogshop_model import CogShopCore, ExternalID
from cognite.powerops.resync.models.base import CDFFile, CDFSequence, DataModel, PowerOpsGraphQLModel

from .graphql_schemas import cogshop1_graphql


class CogShop1Asset(CogShopCore, DataModel, protected_namespaces=()):
    cls_by_container: ClassVar[dict[ContainerId, type[DomainModelApplyCogShop1]]] = {
        ContainerId("cogShop", "ModelTemplate"): cogshop_v1.ModelTemplateApply,
        ContainerId("cogShop", "Mapping"): cogshop_v1.MappingApply,
        ContainerId("cogShop", "FileRef"): cogshop_v1.FileRefApply,
        ContainerId("cogShop", "Transformation"): cogshop_v1.TransformationApply,
        ContainerId("cogShop", "Scenario"): cogshop_v1.ScenarioApply,
        ContainerId("cogShop", "CommandsConfig"): cogshop_v1.CommandsConfigApply,
    }
    graph_ql: ClassVar[PowerOpsGraphQLModel] = cogshop1_graphql
    model_templates: dict[ExternalID, cogshop_v1.ModelTemplateApply] = Field(default_factory=dict)
    mappings: dict[ExternalID, cogshop_v1.MappingApply] = Field(default_factory=dict)
    transformations: dict[ExternalID, cogshop_v1.TransformationApply] = Field(default_factory=dict)
    base_mappings: list[CDFSequence] = Field(default_factory=list)
    output_definitions: list[CDFSequence] = Field(default_factory=list)
    scenarios: dict[ExternalID, cogshop_v1.ScenarioApply] = Field(default_factory=dict)
    commands_configs: dict[ExternalID, cogshop_v1.CommandsConfigApply] = Field(default_factory=dict)

    @field_validator("base_mappings", mode="after")
    def ordering_sequences(cls, value: list[CDFSequence]) -> list[CDFSequence]:
        return sorted(value, key=lambda x: x.external_id)

    @field_validator("model_templates", "mappings", "transformations", mode="after")
    def ordering_dict(cls, value: dict) -> dict:
        return {k: v for k, v in sorted(value.items(), key=lambda x: x[0])}

    @classmethod
    def from_cdf(cls, client: PowerOpsClient, data_set_external_id: str) -> CogShop1Asset:
        cog_shop = client.cog_shop1
        templates = cog_shop.model_template.list(limit=-1, source="resync")
        scenarios = cog_shop.scenario.list(limit=-1, source="resync")

        base_mapping_ids = list(
            {m for t in templates for m in t.base_mappings or []}
            | {m for s in scenarios for m in s.mappings_override or []}
        )

        if base_mapping_ids:
            base_mappings = cog_shop.mapping.retrieve(external_id=base_mapping_ids)
        else:
            base_mappings = []

        if transformation_ids := list({t for m in base_mappings for t in m.transformations or []}):
            transformations = cog_shop.transformation.retrieve(transformation_ids)
        else:
            transformations = []

        if file_ids := list({t.model for t in templates} | {f for s in scenarios for f in s.extra_files or []}):
            files = cog_shop.file_ref.retrieve(file_ids)
        else:
            files = []

        if command_ids := list({s.commands for s in scenarios}):
            commands_configs = cog_shop.commands_config.retrieve(command_ids)
        else:
            commands_configs = []

        transformation_by_id: dict[str, cogshop_v1.TransformationApply] = {
            t.external_id: t.as_apply() for t in transformations
        }

        mappings_by_id: dict[str, cogshop_v1.MappingApply] = {}
        for mapping in base_mappings:
            apply = mapping.as_apply()
            apply.transformations = sorted(
                (transformation_by_id[t] for t in apply.transformations or [] if t in transformation_by_id),
                key=lambda x: x.order,
            )
            mappings_by_id[apply.external_id] = apply

        command_configs_by_id = {commands.external_id: commands.as_apply() for commands in commands_configs}

        file_by_id: dict[str, cogshop_v1.FileRefApply] = {f.external_id: f.as_apply() for f in files}
        model_templates: dict[str, cogshop_v1.ModelTemplateApply] = {}
        for template in templates:
            apply = template.as_apply()
            if apply.model in file_by_id:
                apply.model = file_by_id[apply.model]
            apply.base_mappings = [mappings_by_id[m] for m in apply.base_mappings or [] if m in mappings_by_id]
            model_templates[apply.external_id] = apply

        scenarios_by_id = {}
        for scenario in scenarios:
            apply = scenario.as_apply()
            apply.mappings_override = [mappings_by_id[m] for m in apply.mappings_override or []]
            apply.commands = command_configs_by_id.get(apply.commands)
            apply.model_template = model_templates.get(apply.model_template)
            scenarios_by_id[apply.external_id] = apply

        # There files and sequences are not linked to the model templates
        # (this should be done in the next version of Cogshop).
        watercourse_names = list({t.watercourse for t in model_templates.values()})
        sequence_ids = [
            f"SHOP_{name}{suffix}"
            for name, suffix in product(watercourse_names, ["_base_mapping", "_output_definition"])
        ]
        cdf_client: CogniteClient = client.cdf
        if sequence_ids:
            sequences = cdf_client.sequences.retrieve_multiple(external_ids=sequence_ids)
        else:
            all_sequences = cdf_client.sequences.list(limit=-1, external_id_prefix="SHOP_")
            sequences = [
                s
                for s in all_sequences
                if s.external_id.endswith("_base_mapping") or s.external_id.endswith("_output_definition")
            ]

        base_mappings = [CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_base_mapping")]
        output_definitions = [
            CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_output_definition")
        ]

        files = cdf_client.files.list(limit=-1, source="PowerOps bootstrap", mime_type="text/plain")
        shop_files = [CDFFile(meta=f) for f in files]

        return cls(
            model_templates=model_templates,
            mappings=mappings_by_id,
            transformations=transformation_by_id,
            base_mappings=base_mappings,
            output_definitions=output_definitions,
            shop_files=shop_files,
            scenarios=scenarios_by_id,
            commands_configs=command_configs_by_id,
        )

    def standardize(self) -> None:
        super().standardize()
        self.base_mappings = self.ordering_sequences(self.base_mappings)
        self.output_definitions = self.ordering_sequences(self.output_definitions)
        self.model_templates = self.ordering_dict(self.model_templates)
        self.mappings = self.ordering_dict(self.mappings)
        self.transformations = self.ordering_dict(self.transformations)
        self.scenarios = self.ordering_dict(self.scenarios)
        self.commands_configs = self.ordering_dict(self.commands_configs)
        for scenario in self.scenarios.values():
            scenario.mappings_override = sorted(
                scenario.mappings_override or [], key=lambda x: x if isinstance(x, str) else x.external_id
            )

        for template in self.model_templates.values():
            template.base_mappings = sorted(
                template.base_mappings or [], key=lambda x: x if isinstance(x, str) else x.external_id
            )

        for mapping in self.mappings.values():
            mapping.transformations = sorted(
                mapping.transformations or [], key=lambda x: x[-1] if isinstance(x, str) else x.external_id
            )



from .core import DATAMODEL_ID_TO_RESYNC_NAME, MODELS_BY_NAME, apply, destroy, init, plan, validate
from .core.echo import Echo
from .v2.main import apply2

__all__ = ["apply", "plan", "init", "destroy", "validate", "MODELS_BY_NAME", "Echo", "DATAMODEL_ID_TO_RESYNC_NAME"]



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._bid_method import BidMethod, BidMethodWrite


__all__ = [
    "PriceArea",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
]


PriceAreaTextFields = Literal["name", "timezone", "main_scenario", "price_scenarios"]
PriceAreaFields = Literal["name", "timezone", "main_scenario", "price_scenarios"]

_PRICEAREA_PROPERTIES_BY_FIELD = {
    "name": "name",
    "timezone": "timezone",
    "main_scenario": "mainScenario",
    "price_scenarios": "priceScenarios",
}


class PriceArea(DomainModel):
    """This represents the reading version of price area.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: The name field.
        default_method: The default method field.
        timezone: The timezone field.
        main_scenario: The main scenario field.
        price_scenarios: The price scenario field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "PriceArea")
    name: str
    default_method: Union[BidMethod, str, dm.NodeId, None] = Field(None, repr=False, alias="defaultMethod")
    timezone: str
    main_scenario: Union[TimeSeries, str, None] = Field(None, alias="mainScenario")
    price_scenarios: Union[list[TimeSeries], list[str], None] = Field(None, alias="priceScenarios")

    def as_write(self) -> PriceAreaWrite:
        """Convert this read version of price area to the writing version."""
        return PriceAreaWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            default_method=(
                self.default_method.as_write() if isinstance(self.default_method, DomainModel) else self.default_method
            ),
            timezone=self.timezone,
            main_scenario=self.main_scenario,
            price_scenarios=self.price_scenarios,
        )

    def as_apply(self) -> PriceAreaWrite:
        """Convert this read version of price area to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaWrite(DomainModelWrite):
    """This represents the writing version of price area.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: The name field.
        default_method: The default method field.
        timezone: The timezone field.
        main_scenario: The main scenario field.
        price_scenarios: The price scenario field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "PriceArea")
    name: str
    default_method: Union[BidMethodWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="defaultMethod")
    timezone: str
    main_scenario: Union[TimeSeries, str, None] = Field(None, alias="mainScenario")
    price_scenarios: Union[list[TimeSeries], list[str], None] = Field(None, alias="priceScenarios")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(PriceArea, dm.ViewId("power-ops-day-ahead-bid", "PriceArea", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.default_method is not None:
            properties["defaultMethod"] = {
                "space": self.space if isinstance(self.default_method, str) else self.default_method.space,
                "externalId": (
                    self.default_method if isinstance(self.default_method, str) else self.default_method.external_id
                ),
            }

        if self.timezone is not None:
            properties["timezone"] = self.timezone

        if self.main_scenario is not None or write_none:
            if isinstance(self.main_scenario, str) or self.main_scenario is None:
                properties["mainScenario"] = self.main_scenario
            else:
                properties["mainScenario"] = self.main_scenario.external_id

        if self.price_scenarios is not None or write_none:
            properties["priceScenarios"] = [
                value if isinstance(value, str) else value.external_id for value in self.price_scenarios or []
            ] or None

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.default_method, DomainModelWrite):
            other_resources = self.default_method._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.main_scenario, CogniteTimeSeries):
            resources.time_series.append(self.main_scenario)

        if isinstance(self.price_scenarios, CogniteTimeSeries):
            resources.time_series.append(self.price_scenarios)

        return resources


class PriceAreaApply(PriceAreaWrite):
    def __new__(cls, *args, **kwargs) -> PriceAreaApply:
        warnings.warn(
            "PriceAreaApply is deprecated and will be removed in v1.0. Use PriceAreaWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceArea.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PriceAreaList(DomainModelList[PriceArea]):
    """List of price areas in the read version."""

    _INSTANCE = PriceArea

    def as_write(self) -> PriceAreaWriteList:
        """Convert these read versions of price area to the writing versions."""
        return PriceAreaWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceAreaWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaWriteList(DomainModelWriteList[PriceAreaWrite]):
    """List of price areas in the writing version."""

    _INSTANCE = PriceAreaWrite


class PriceAreaApplyList(PriceAreaWriteList): ...


def _create_price_area_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    timezone: str | list[str] | None = None,
    timezone_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if default_method and isinstance(default_method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultMethod"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": default_method},
            )
        )
    if default_method and isinstance(default_method, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultMethod"),
                value={"space": default_method[0], "externalId": default_method[1]},
            )
        )
    if default_method and isinstance(default_method, list) and isinstance(default_method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultMethod"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in default_method],
            )
        )
    if default_method and isinstance(default_method, list) and isinstance(default_method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultMethod"),
                values=[{"space": item[0], "externalId": item[1]} for item in default_method],
            )
        )
    if isinstance(timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timezone"), value=timezone))
    if timezone and isinstance(timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timezone"), values=timezone))
    if timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timezone"), value=timezone_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
